# Copyright (C) 2009 Luo Jinghua <sunmoon1997@gmail.com>.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import pygtk
pygtk.require('2.0')

import sys
import gobject
import gtk
import totem

class PPSVideoList (gtk.TreeView):
    __gproperties__ = {
        'title-column' : (gobject.TYPE_INT, "title-column",
                          "title-column", -1, gobject.G_MAXINT, -1,
                          gobject.PARAM_READWRITE),
        'tooltip-column' : (gobject.TYPE_INT, 'tooltip-column',
                            'tooltip-column', -1, gobject.G_MAXINT, -1,
                            gobject.PARAM_READWRITE),
        'mrl-column' : (gobject.TYPE_INT, 'mrl-column', 'mrl-column',
                        -1, gobject.G_MAXINT, -1, gobject.PARAM_READWRITE),
        "totem": (gobject.TYPE_OBJECT, "Totem",
                  "Totem", gobject.PARAM_READWRITE),
        "plugin": (gobject.TYPE_OBJECT, "plugin",
                  "plugin", gobject.PARAM_READWRITE),
    }
    __gsignals__ = { 'starting-video': (gobject.SIGNAL_RUN_LAST, gobject.TYPE_BOOLEAN,
                                        (gobject.TYPE_OBJECT, gobject.TYPE_PYOBJECT )),
    }

    __gtype_name__ = 'PPSVideoList'

    def __init__ (self):
        self.plugin = None
        self.totem = None
        self.builder = None
        gtk.TreeView.__init__ (self)

        self.set_property ("has-tooltip", True)
        self.connect ("row-activated", self.row_activated_cb)
        self.connect ("query-tooltip", self.query_tooltip_cb)
        self.connect ("button-press-event", self.button_pressed_cb)
        self.connect ("popup-menu", self.popup_menu_cb)

        selection = self.get_selection ()
        selection.connect ("changed", self.selection_changed_cb)
        selection.set_mode (gtk.SELECTION_MULTIPLE)

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def row_activated_cb (self, widget, path, column):
        #mrl_column =  widget.get_property ("mrl-column")
        if self.get_property ("mrl-column") == -1:
            return

        play_video = self.emit ("starting-video", self, path)
        if not play_video:
            return

        model = widget.get_model ()
        iter = model.get_iter (path)
        mrl = model.get (iter, self.get_property ("mrl-column"))[0]
        title = model.get (iter, self.get_property ("title-column"))[0]
        self.totem.add_to_playlist_and_play (mrl, title, False)

    def query_tooltip_cb (self, widget, x, y, keyboard_mode, tooltip):
        if not widget.get_tooltip_context (x, y, keyboard_mode):
            return False

        model, path, iter = widget.get_tooltip_context (x, y, keyboard_mode)
        mrl_column =  self.get_property ("mrl-column")
        if mrl_column == -1:
            text = model.get (iter, self.get_property ("tooltip-column"))[0]
            tooltip.set_text (text)
        else:
            text = model.get (iter, self.get_property ("tooltip-column"))[0]
            text += '\n' + model.get (iter, self.get_property ("mrl-column"))[0]
            tooltip.set_text (text)
        widget.set_tooltip_row (tooltip, path)
        return True

    def button_pressed_cb (self, widget, event):
        pass

    def selection_changed_cb (self, widget):
        self.trigger_tooltip_query ()

    def popup_menu_cb (self, treeview, widget):
        return False

    def get_ui_manager (self):
        return self.ui_manager

gobject.type_register (PPSVideoList)
