import totem
import gobject, gtk, gconf
gobject.threads_init()
import urllib
import httplib
import atom
import threading
import time
import re
import os
import random

import PPSVideoList

from xml.dom import minidom

def xml_node_value (node):
	return node.firstChild.nodeValue

def xml_child_value (node, id):
	child = node.getElementsByTagName(id)[0].firstChild;
	if child:
		return child.nodeValue
	return None

class PPSClass:
	def __str__(self):
		return str (self.id) + ' ' + self.title

class PPSSubClass:
	def __str__(self):
		return str (self.subid) + ' ' + self.title

class PPSFile:
	def __str__(self):
		return str (self.id) + ' ' + self.url

class PPSFilm:
	def __str__(self):
		return str (self.id) + ' ' + self.title

def parseclass (node):
	result = PPSClass()
	result.id = int(xml_child_value (node, 'ID'))
	result.title = str(xml_child_value (node, 'Title'))
	result.type = int(xml_child_value (node, 'Type'))
	result.contentnum = int(xml_child_value (node, 'ContentNum'))
	result.subclasses = []
	result.films = []
	return result

def parsesubclass (node):
	result = PPSSubClass ()
	result.subid = int(xml_child_value (node, 'SubID') or 0)
	result.title = str(xml_child_value (node, 'Title'))
	result.filmtitle = str(xml_child_value (node, 'FilmTitle'))
	result.director = str(xml_child_value (node, 'Dirt'))
	result.actor = str(xml_child_value (node, 'Actor'))
	result.bigimage = str(xml_child_value (node, 'BIUrl'))
	result.smallimage = str(xml_child_value (node, 'SIUrl'))
	result.contentnum = int(xml_child_value (node, 'ContentNum'))
	result.contents = []
	result.films = []
	return result

def parsefile (node):
	result = PPSFile ()
	result.id = int(xml_child_value (node, 'ID'))
	result.ci = int(xml_child_value (node, 'CI'))
	result.size = int(xml_child_value (node, 'Size'))
	result.format = str(xml_child_value (node, 'Form'))
	result.rate = int(xml_child_value (node, 'Rate'))
	result.url = str(xml_child_value (node, 'DUrl'))
	result.vipflag = int(xml_child_value (node, 'VF'))
	return result

def parsefilm (node):
	result = PPSFilm ()
	result.id = int(xml_child_value (node, 'ID'))
	result.title = str(xml_child_value (node, 'Title'))
	result.director = str(xml_child_value (node, 'Dirt'))
	result.actor = str(xml_child_value (node, 'Actor'))
	result.area = str(xml_child_value (node, 'Area'))
	result.size = int(xml_child_value (node, 'Size'))
	result.pubtime = str(xml_child_value (node, 'PT'))
	result.length = str(xml_child_value (node, 'LT'))
	result.lang = str(xml_child_value (node, 'Lang'))
	result.score = float(xml_child_value (node, 'Score'))
	result.desc = str(xml_child_value (node, 'Desc'))
	result.longdesc = str(xml_child_value (node, 'Ldesc'))
	result.bigimage = str(xml_child_value (node, 'BIUrl'))
	result.smallimage = str(xml_child_value (node, 'SIUrl'))
	result.cn = int(xml_child_value (node, 'CN'))
	result.files = []
	result.pixbuf = None

	down = node.getElementsByTagName ('Down')
	if not down:
		return
	files = down[0].getElementsByTagName ('Url')
	for f in files:
		result.files += [parsefile (f)]
	return result

class DownloadThread (threading.Thread):
	def __init__ (self, ppstream, url, what = None, *args, **kwargs):
		self.ppstream = ppstream
		self.url = url
		#self.treeview_name = treeview_name
		self.what = what
		self.args = args
		self.kwargs = kwargs
		self._done = False
		self._lock = threading.Lock ()
		threading.Thread.__init__ (self)

	def run (self):
		try:
			res = urllib.urlopen(self.url).read()
		except Exception, e:
			print e
			res = None
		gobject.idle_add (self.publish_results, res)

	def publish_results(self, res):
		self._lock.acquire (True)
		self.ppstream.fetched (self.what, res, *self.args, **self.kwargs)
		self._done = True
		self._lock.release ()
		return False

	@property
	def done (self):
		""" Thread-safe property to know whether the query is done or not """
		self._lock.acquire (True)
		res = self._done
		self._lock.release ()
		return res

class CallbackThread (threading.Thread):
	def __init__ (self, callback, *args, **kwargs):
		self.callback = callback
		self.args = args
		self.kwargs = kwargs
		threading.Thread.__init__ (self)

	def run (self):
		res = self.callback (*self.args, **self.kwargs)
		while res == True:
			res = self.callback (*self.args, **self.kwargs)

class RetrieveImageTask:
	pass

class RetrieveImage (threading.Thread):
	def __init__ (self):
		self.queue = []
		self._cache = {}
		self._lock = threading.Lock ()
		self._done = False
		threading.Thread.__init__ (self)

	def retrieve (self, url, callback, *args):
		task = RetrieveImageTask ()
		task.url = url
		task.callback = callback
		task.args = args
		self._lock.acquire ()
		self.queue.append (task)
		self._lock.release ()

	def cancel_tasks (self):
		self._lock.acquire ()
		self.queue = []
		self._cache = {}
		self._lock.release ()

	def cancel (self):
		self._done = True
		self._lock.acquire ()
		self.queue = []
		self._cache = {}
		self._lock.release ()

	def load_pixbuf (self, url):
		try:
			filename, headers = urllib.urlretrieve (url)
		except IOError:
			return None

		try:
			pixbuf = gtk.gdk.pixbuf_new_from_file (filename)
		except gobject.GError:
			return None
		return pixbuf

	def run (self):
		import time
		while not self.done:
			task = None
			self._lock.acquire ()
			if len (self.queue):
				task = self.queue.pop (0)
			self._lock.release ()

			if not task:
				time.sleep (0.1)
				continue

			if task.url in self.cache:
				pixbuf = self.cache [task.url]
			else:
				pixbuf = self.load_pixbuf (task.url)
				if pixbuf:
					self.cache [task.url] = pixbuf
			if not pixbuf:
				continue

			gobject.idle_add (self.publish_result, pixbuf, task)

	def publish_result (self, pixbuf, task):
		task.callback (pixbuf, *task.args)
		return False

	@property
	def done (self):
		self._lock.acquire (True)
		res = self._done
		self._lock.release ()
		return res

	@property
	def cache (self):
		self._lock.acquire (True)
		res = self._cache
		self._lock.release ()
		return res

class PPStream (totem.Plugin):
	def __init__ (self):
		totem.Plugin.__init__ (self)
		self.debug = False
		self.gstreamer_plugins_present = True

		"""Search counters (per search type)"""
		self.in_search = {}
		self.search_token = {} # Used as an ID for a search thread

		self.max_results = 20
		self.button_down = False

		self.search_terms = ""
		self.ppstream_id = ""

		self.start_index = {}
		self.results = {} # This is just the number of results from the last pagination query
		self.entry = {}

		self.current_treeview_name = ""
		self.notebook_pages = []

		self.vadjust = {}
		self.liststore = {}
		self.treeview = {}

		self.classes = []
		self.subclassindex = 0

	def activate (self, totem_object):
		"""Check for the availability of the ppssrc GStreamer plugins"""
		bvw_name = totem_object.get_video_widget_backend_name ()

		if bvw_name.find ("GStreamer") != -1:
			try:
				import pygst
				pygst.require ("0.10")
				import gst

				registry = gst.registry_get_default ()
				if registry.find_plugin ("pps") == None:
					"""This means an error will be displayed when they try to play anything"""
					self.gstreamer_plugins_present = False
			except ImportError:
				"""Do nothing; either it's using xine or python-gstreamer isn't installed"""

		"""Continue loading the plugin as before"""
		self.builder = self.load_interface ("ppstream.ui", True, totem_object.get_main_window (), self)
		self.totem = totem_object

		self.search_entry = self.builder.get_object ("pps_search_entry")
		self.search_entry.connect ("activate", self.on_search_entry_activated)
		self.search_button = self.builder.get_object ("pps_search_button")
		self.search_button.connect ("clicked", self.on_search_button_clicked)
		self.progress_bar = self.builder.get_object ("pps_progress_bar")

		self.notebook = self.builder.get_object ("pps_notebook")
		self.notebook.connect ("switch-page", self.on_notebook_page_changed)

		self.notebook_pages = ["search", "related"]
		self.current_treeview_name = "search"
		self.setup_categories ()
		self.setup_films ()

		self.vbox = self.builder.get_object ("pps_vbox")
		self.vbox.show_all ()
		totem_object.add_sidebar_page ("ppstream", _("PPStream"), self.vbox)

		self.fetch_classes ()

		self.retrieveimage = RetrieveImage ()
		self.retrieveimage.start ()

	def deactivate (self, totem):
		totem.remove_sidebar_page ("ppstream")
		self.classes = []
		self.retrieveimage.cancel ()
		self.retrieveimage.join ()

	def fetch_classes (self, force = False):
		if force:
			self.classes = []
		if self.classes == []:
			self.subclassindex = 0
			url = 'http://playlist.pps.tv/classfortv.php'
			thread = DownloadThread (self, url, "class")
			thread.start()

	def fetched (self, what, res, *args):
		if what == "class":
			if res:
				self.parseclass (res)
		elif what == "subclass":
			### parse the xml
			self.parsesubclass (res)

			### fill categories view
			self.fill_categories (self.classes[self.subclassindex].id)

			### fetch next subclass
			self.fetchnextsubclass ()
		elif what == 'film':
			result = self.parsefilm (res, *args)
			if result:
				self.fill_films (result)
				self.fetchfilms (*args)

	def parseclass (self, res):
		xmldoc = minidom.parseString (res)
		ch = xmldoc.getElementsByTagName ('PPSChannels')
		if not ch:
			return
		ch = ch[0]
		cls = ch.getElementsByTagName ('Class')
		for cl in cls:
			self.classes += [parseclass (cl)]

		if cls:
			self.fetchsubclass (0)

	def fetchsubclass (self, index):
		self.subclassindex = index
		if self.classes[index].type == 1:
			self.fetchnextsubclass ()
			return
		url = 'http://playlist.pps.tv/subclassfortv.php?class='
		classid = str(self.classes[index].id);
		thread = DownloadThread (self, url + classid, "subclass")
		thread.start()

	def fetchnextsubclass (self):
		if self.subclassindex < len(self.classes) - 1:
			self.fetchsubclass (self.subclassindex + 1)

	def parsesubclass (self, res):
		if not res:
			return
		xmldoc = minidom.parseString (res)
		clses = xmldoc.getElementsByTagName ('PPSSClasses')
		if not clses:
			return
		cls = clses[0].getElementsByTagName ('SubClass')
		curclass = self.classes[self.subclassindex]
		for cl in cls:
			curclass.subclasses += [parsesubclass (cl)]

	def classindex (self, clsid):
		i = 0
		for cls in self.classes:
			if cls.id == clsid:
				return i
			i += 1
		return -1

	def getsubclassindex (self, clsindex, subclsid):
		i = 0
		for cls in self.classes[clsindex].subclasses:
			if cls.subid == subclsid:
				return i
			i += 1
		return -1

	def parsefilm (self, res, clsid, subclsid):
		if not res:
			return []
		xmldoc = minidom.parseString (res)
		data = xmldoc.getElementsByTagName ('FilmData')
		if not data:
			return []
		films = data[0].getElementsByTagName ('Film')
		if not films:
			return []
		cls, subcls = self.getclasspair (clsid, subclsid)
		if not subcls:
			subcls = cls

		result = []
		for film in films:
			result += [parsefilm (film)]

		subcls.films += result
		return result

	def getclasspair (self, clsid, subclsid):
		clsindex = self.classindex (clsid)
		cls = self.classes[clsindex]
		if cls.type == 0:
			subclsindex = self.getsubclassindex (clsindex, subclsid)

		if cls.type == 0:
			subcls = cls.subclasses[subclsindex]
		else:
			subcls = None
		return cls, subcls

	def fetchfilms (self, clsid, subclsid):
		cls, subcls = self.getclasspair (clsid, subclsid)
		if cls.type == 1:
			subcls = cls
		if len (subcls.films) >= subcls.contentnum:
			return

		url = 'http://playlist.pps.tv/channelsfortv.php?'
		url += 'class=' + str(clsid)
		if cls.type == 0:
			url += '&subclass=' + str(subclsid)
		if not len (subcls.films):
			subcls.page = 0
		else:
			subcls.page += 1
		#print len(subcls.films), subcls.films, subcls.contentnum
		url += '&page=' + str(subcls.page + 1)
		#print 'fetchfilm url = ', url
		thread = DownloadThread (self, url, "film", clsid, subclsid)
		thread.start()

	def on_cat_treeview_row_activated(self, treeview, path, view_column, data=None):
		model, rows = treeview.get_selection ().get_selected_rows ()
		iter = model.get_iter (rows[0])
		title = model.get_value (iter, 3)
		classid = model.get_value (iter, 0)
		subclassid = model.get_value (iter, 2)

		print title, classid, subclassid

		self.retrieveimage.cancel_tasks ()
		treeview_name = 'related'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		treeview.clear ()

		self.show_films (classid, subclassid)

	def setup_categories (self):
		treeview_name = 'search'
		treeview = self.builder.get_object ("pps_treeview_" + treeview_name)
		renderer = gtk.CellRendererText()
		renderer.set_property("xalign", 0.0)

		column = gtk.TreeViewColumn(_("Name"), renderer, text = 3)
		column.set_clickable(True)
		treeview.append_column(column)

		column = gtk.TreeViewColumn(_("Number"), renderer, text = 4)
		treeview.append_column(column)

		self.vadjust[treeview_name] = treeview.get_vadjustment ()
		self.vadjust[treeview_name].connect ("value-changed", self.on_value_changed)
		vscroll = self.builder.get_object ("pps_scrolled_window_" + treeview_name).get_vscrollbar ()
		vscroll.connect ("button-press-event", self.on_button_press_event)
		vscroll.connect ("button-release-event", self.on_button_release_event)

		self.liststore[treeview_name] = self.builder.get_object ("pps_liststore_" + treeview_name)
		self.treeview[treeview_name] = treeview
		treeview.set_model (self.liststore[treeview_name])

		treeview.connect("row_activated", self.on_cat_treeview_row_activated)

	def fill_all_categories (self):
		treeview_name = 'search'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		self.liststore[treeview_name].clear()
		for cls in self.classes:
			iter = treeview.append(None)
			treeview.set(iter,
				     0, cls.id,
				     1, cls.type,
				     2, 0,
				     3, cls.title,
				     4, cls.contentnum)

			# add children
			for subcls in cls.subclasses:
				child_iter = treeview.append(iter);
				treeview.set(child_iter,
					     0, cls.id,
					     1, cls.type,
					     2, subcls.subid,
					     3, subcls.title,
					     4, subcls.contentnum)

	def fill_categories (self, clsid):
		treeview_name = 'search'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		for cls in self.classes:
			if cls.id != clsid:
				continue

			iter = treeview.append(None)
			treeview.set(iter,
				     0, cls.id,
				     1, cls.type,
				     2, 0,
				     3, cls.title,
				     4, cls.contentnum)

			# add children
			for subcls in cls.subclasses:
				child_iter = treeview.append(iter);
				treeview.set(child_iter,
					     0, cls.id,
					     1, cls.type,
					     2, subcls.subid,
					     3, subcls.title,
					     4, subcls.contentnum)
			break

	def gettipfromfilm (self, film):
		tip = film.title
		tip += '\n' + _("Director: ") + film.director
		tip += '\n' + _("Actor: ") + film.actor
		tip += '\n' + _("Area: ") + film.area
		tip += '\n' + _("Size: ") + str(film.size) + _('MB')
		tip += '\n' + _("Pubtime: ") + film.pubtime
		tip += '\n' + _("Length: ") + str(film.length) + _('Min')
		tip += '\n' + _("Score: ") + str(film.score)
		tip += '\n' + _("Description: ") + film.desc
		tip += '\n' + _("Long Description: ") + film.longdesc

		return tip

	def retrieveimage_cb (self, pixbuf, film, iter):
		treeview_name = 'related'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		film.pixbuf = pixbuf

		try:
			treeview.set (iter, 0, pixbuf)
		except:
			pass


	def storelist_append_file (self, treeview, film, f):
		tip = self.gettipfromfilm (film)
		it = treeview.append ([film.pixbuf, film.title, f.url, tip])
		if not film.pixbuf and film.smallimage:
			self.retrieveimage.retrieve (film.smallimage, self.retrieveimage_cb,
								     film, it)
	def fill_all_films (self, clsid, subclsid):
		cls, subcls = self.getclasspair (clsid, subclsid)
		if cls.type == 1:
			subcls = cls
		treeview_name = 'related'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		treeview.clear ()
		for film in subcls.films:
			for f in film.files:
				self.storelist_append_file (treeview, film, f)

	def fill_films (self, films):
		treeview_name = 'related'
		treeview = self.builder.get_object ("pps_liststore_" + treeview_name)
		for film in films:
			for f in film.files:
				self.storelist_append_file (treeview, film, f)

	def show_films (self, clsid, subclsid):
		### FIXME what to do when the subclasses is empty?
		cls, subcls = self.getclasspair (clsid, subclsid)
		if cls.type == 1:
			subcls = cls
		if len (subcls.films) >= subcls.contentnum:
			self.fill_all_films (clsid, subclsid)
			return
		self.fetchfilms (clsid, subclsid)

	def setup_films (self):
		treeview_name = 'related'
		self.start_index[treeview_name] = 1
		self.results[treeview_name] = 0
		self.entry[treeview_name] = None
		self.in_search[treeview_name] = False

		"""This is done here rather than in the UI file,
		because UI files parsed in C and GObjects created in
		Python apparently don't mix."""
		renderer = totem.CellRendererVideo (use_placeholder = True)
		treeview = self.builder.get_object ("pps_treeview_" + treeview_name)
		treeview.set_property ("totem", self.totem)
		treeview.connect ("row-activated", self.on_row_activated)
		treeview.connect_after ("starting-video", self.on_starting_video)
		treeview.insert_column_with_attributes (0, _("Videos"), renderer, thumbnail=0, title=1)

		self.vadjust[treeview_name] = treeview.get_vadjustment ()
		self.vadjust[treeview_name].connect ("value-changed", self.on_value_changed)
		vscroll = self.builder.get_object ("pps_scrolled_window_" + treeview_name).get_vscrollbar ()
		vscroll.connect ("button-press-event", self.on_button_press_event)
		vscroll.connect ("button-release-event", self.on_button_release_event)

		self.liststore[treeview_name] = self.builder.get_object ("pps_liststore_" + treeview_name)
		self.treeview[treeview_name] = treeview
		treeview.set_model (self.liststore[treeview_name])

	def on_notebook_page_changed (self, notebook, notebook_page, page_num):
		self.current_treeview_name = self.notebook_pages[page_num]

	def on_row_activated (self, treeview, path, column):
		if self.debug:
			print "Activating row"

		model, rows = treeview.get_selection ().get_selected_rows ()
		iter = model.get_iter (rows[0])
		ppstream_id = model.get_value (iter, 3)

		if self.debug:
			print "Done activating row"

	def on_starting_video (self, treeview, path, user_data):
		"""Display an error if the required GStreamer plugins aren't installed"""
		if self.gstreamer_plugins_present == False:
			self.totem.interface_error_with_link (_("Totem cannot play this type of media (%s) because you do not have the appropriate plugins to handle it.") % _("Ppstream"),
							      _("Please install the necessary plugins and restart Totem to be able to play this media."),
							      "http://www.gnome.org/projects/totem/#codecs",
							      _("More information about media plugins"),
							      self.totem.get_main_window ())
			return False

		return True

	def on_button_press_event (self, widget, event):
		self.button_down = True

	def on_button_release_event (self, widget, event):
		self.button_down = False
		self.on_value_changed (self.vadjust[self.current_treeview_name])

	def on_value_changed (self, adjustment):
		"""Load more results when we get near the bottom of the treeview"""
		pass

	def _clear_ui (self, treeview_name):
		"""Revert the cursor"""
		window = self.vbox.window
		window.set_cursor (None)
		self.progress_bar.set_fraction (0.0)
		self.progress_bar.set_text ("")

		return False

	def on_search_button_clicked (self, button):
		search_terms = self.search_entry.get_text ()

		if self.debug:
			print "Searching for \"" + search_terms + "\""

		"""Focus the "Search" tab"""
		self.notebook.set_current_page (self.notebook_pages.index ("search"))

	def on_search_entry_activated (self, entry):
		self.search_button.clicked ()

